#include "../defs.h"
#include "../errors.h"
#include "../Lock/lock.h"
#include "../mountrec.h"
#include "readwrite.h"
#include "../Volume/volinfo.h"
#include "../Volume/volbmap.h"
#include "../LibFuncs/macmount.h"
#include "../DevSupport/partition.h"
#include <stdio.h>
#include <sys/file.h>
#include <sys/uio.h>
#ifdef I486
#include <sys/types.h>
#endif
#include <ctype.h>
#include <sysent.h>

/*
 * Call first to open a new device and allocate a state record for the 
 * device.
 *
 */
#ifdef USER
int initreadwrite(char *devname, int partnum, MOUNT_RECORD *mr)
{
  int rc;

  if ((mr->sr.diskfd=open(devname,O_RDWR,0))==-1) {
    return INITREADWRITEFAILED;
  }

  mr->sr.validflag=VALIDVAL;
  mr->sr.maxwritesbeforemrflush=MAXWRITESBEFOREMRFLUSH;
  mr->sr.currentnumwrites=0;

  if ((rc=getpartinfo(mr, partnum))) {
    mr->sr.validflag=0;
    return rc;
  }

  return 0;
}
#endif

int closereadwrite(MOUNT_RECORD *mr)
{
#ifdef USER
  if ((close(mr->sr.diskfd))==-1) {
    return CLOSEREADWRITEFAILED;
  }
#endif

  mr->sr.validflag=!VALIDVAL;

  return 0;
}


#ifdef USER
int readabsolutesector(MOUNT_RECORD *mr, SECTNUMTYPE sect, char *buf)
{
  struct staterec *what;

  what=&(mr->sr);

  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  if (lseek(what->diskfd, sect * what->sectsize, L_SET)==-1)
    return READSECTFAILED;

  if (read(what->diskfd, buf, what->sectsize)<what->sectsize)
    return READSECTFAILED;

  return 0;

}
#endif

/*
 *
 * Returns the number of sectors read, or a negative number if an error
 * occurred.
 *
 */
#ifdef USER
int readsects(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  long numread;
  struct staterec *what;

  what=&(mr->sr);

  sect+=what->sectoffsettopartition;

  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  if (lseek(what->diskfd, sect*what->sectsize, L_SET)==-1)
    return READSECTFAILED;

  if ((numread=read(what->diskfd, buf, num*what->sectsize))<what->sectsize)
    return READSECTFAILED;

  if (numread%what->sectsize)
    return READSECTFAILED;

  return numread/what->sectsize;

}
#endif

/*
 *
 * Returns the number of sectors written, or a negative number if an error
 * occurred.
 *
 */
#ifdef USER
int writesects(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  long numwritten;
  struct staterec *what;
  int rc;

  what=&(mr->sr);

  sect+=what->sectoffsettopartition;

  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  if (lseek(what->diskfd, sect*what->sectsize, L_SET)==-1)
    return WRITESECTFAILED;

  if ((numwritten=write(what->diskfd, buf, num*what->sectsize))<what->sectsize)
    return WRITESECTFAILED;

  if (numwritten%what->sectsize)
    return WRITESECTFAILED;

  if (what->currentnumwrites>=what->maxwritesbeforemrflush) {
    if ((rc=syncmountrecord(mr))) {
      return rc;
    } else {
      what->currentnumwrites=0;
    }
  } else {
    ++(what->currentnumwrites);
  }

  return numwritten/what->sectsize;

}
#endif

/*
 *
 * Returns the number of sectors written, or a negative number if an error
 * occurred.
 *
 */
#ifdef USER
int writesectsunbuf(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  long numwritten;
  struct staterec *what;

  what=&(mr->sr);

  sect+=what->sectoffsettopartition;


  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  if (lseek(what->diskfd, sect*what->sectsize, L_SET)==-1)
    return WRITESECTUNBUFFAILED;

  if ((numwritten=write(what->diskfd, buf, num*what->sectsize))<what->sectsize)
    return WRITESECTUNBUFFAILED;

  if (numwritten%what->sectsize)
    return WRITESECTUNBUFFAILED;

  return numwritten/what->sectsize;

}
#endif










