#include "../defs.h"
#include "../errors.h"
#include "../Lock/lock.h"
#include "../mountrec.h"
#include "readwrite_vfs.h"
#include "../Volume/volinfo.h"
#include "../Volume/volbmap.h"
#include "../LibFuncs/macmount_vfs.h"
#include "../DevSupport/partition.h"
#include <stdio.h>
#include <sys/file.h>
#include <sys/uio.h>
#ifdef I486
#include <sys/types.h>
#endif
#include <ctype.h>
#include <sysent.h>

/*
 * Call first to open a new device and allocate a state record for the 
 * device.
 *
 */
int initreadwrite(struct vnode *device, int partnum, MOUNT_RECORD *mr)
{
  int rc;

  if (device==NULL)
    return INITREADWRITEFAILED;

  mr->sr.device=device;
  mr->sr.validflag=VALIDVAL;
  mr->sr.maxwritesbeforemrflush=MAXWRITESBEFOREMRFLUSH;
  mr->sr.currentnumwrites=0;


  if ((rc=getpartinfo(mr,partnum))) {
    mr->sr.validflag=0;
    return rc;
  }

  return 0;
}


 
int closereadwrite(MOUNT_RECORD *mr)
{
  mr->sr.validflag=!VALIDVAL;

  return 0;
}


int readabsolutesector(MOUNT_RECORD *mr, SECTNUMTYPE sect, char *buf)
{
  struct staterec *what;
  struct buf *bp;

  what=&(mr->sr);

  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  bp=NULL;
  bp=bread(what->device, sect, what->sectsize);
  if (bp->b_error)
    return READSECTFAILED;
  bcopy(bp->b_un.b_addr,buf,what->sectsize);
  brelse(bp);
  
  return 0;

}




/*
 *
 * Returns the number of sectors read, or a negative number if an error
 * occurred.
 *
 */
int readsects(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  struct staterec *what;
  SECTNUMTYPE cursect;
  struct buf *bp;
  BYTE *bufptr;

  what=&(mr->sr);

  sect+=what->sectoffsettopartition;

  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  bufptr=buf;

  for (cursect=sect;cursect<(sect+num);cursect++, bufptr+=what->sectsize) {
    bp=NULL;
    bp=bread(what->device, cursect, what->sectsize);
    if (bp->b_error)
      return READSECTFAILED;
    bcopy(bp->b_un.b_addr,bufptr,what->sectsize);
    brelse(bp);
  }
  
  return num;

}



/*
 *
 * Returns the number of sectors written, or a negative number if an error
 * occurred.
 *
 */
int writesects(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  struct staterec *what;
  int rc;
  SECTNUMTYPE cursect;
  struct buf *bp;
  BYTE *bufptr;

  what=&(mr->sr);

  sect+=what->sectoffsettopartition;


  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;

  
  bufptr=buf;

  for (cursect=sect;cursect<(sect+num);cursect++, bufptr+=what->sectsize) {
    bp=NULL;
    bp=getblk(what->device, cursect, what->sectsize);
    if (bp->b_error)
      return WRITESECTFAILED;
    bcopy(bufptr, bp->b_un.b_addr, what->sectsize);
    bdwrite(bp);
  }

  if (what->currentnumwrites>=what->maxwritesbeforemrflush) {
    if ((rc=syncmountrecord(mr))) {
      return rc;
    } else {
      what->currentnumwrites=0;
    }
  } else {
    what->currentnumwrites+=num;
  }

  return num;

}

/*
 *
 * Returns the number of sectors written, or a negative number if an error
 * occurred.
 *
 */
int writesectsunbuf(MOUNT_RECORD *mr, SECTNUMTYPE sect, int num, char *buf)
{
  struct staterec *what;
  SECTNUMTYPE cursect;
  BYTE *bufptr;
  struct buf *bp;


  what=&(mr->sr);

  sect+=what->sectoffsettopartition;


  if (what->validflag!=VALIDVAL)
    return INVALIDSTATEREC;
  
  bufptr=buf;

  for (cursect=sect;cursect<(sect+num);cursect++, bufptr+=what->sectsize) {
    bp=NULL;
    bp=getblk(what->device, cursect, what->sectsize);
    if (bp->b_error)
      return WRITESECTFAILED;
    bcopy(bufptr, bp->b_un.b_addr, what->sectsize);
    bwrite(bp);
  }


  return num;

}










