#include <stdio.h>
#include "../Includes/macfs_user.h"
#include <ctype.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <stdlib.h>
#include <sysent.h>
#include <time.h>

#define NOCORRECTNESS

#define MAXITEMS 500
#define NUMWRITES 1000
#define BLOCKSIZ 1000

#define ISGONE 0
#define ISFOLD 1
#define ISFILE 2

struct thing {
  char *pn;
  int whatisit;
} genedpathname[MAXITEMS+1];
 

int maxentry;

MOUNT_RECORD mr;

int main(int argc, char *argv[])
{
  int fail;
  MFILE mfile;
  int fd;
  int b;
  BYTE buf[BLOCKSIZ];
  LONGWORD bytesread, byteswritten, actualbs;
  char *devname, *partnum;
  char fn[MAXFNLENGTH+1];
  int pathlen;
  int rc;
  int i;
  int item;
  int curbs;
  time_t starttime, endtime, gentime;

  devname=getenv("MACDEVICE");
  partnum=getenv("MACPART");


  if (devname==NULL || partnum==NULL) {
    puts("Please set the environment variables MACDEVICE and MACPART before");
    puts("running any mac utility.  MACDEVICE is like /dev/rz4h or /dev/fd0");
    exit(1);
  }

  for (i=0;i<BLOCKSIZ; i++) 
    buf[i]='Z';


  starttime=time(NULL);
  

  rc=macmount(devname, atoi(partnum), &mr);



  if (rc) {
    puts("macmount failed");
    exit(2);
  }

  initopencatrectable();


  genedpathname[0].pn="";
  genedpathname[0].whatisit=ISFOLD;
  
  maxentry=1;

  printf("Now Generating %d files and folders\n", MAXITEMS);

  starttime=time(NULL);

  for (i=1;i<=MAXITEMS;i++) {
    endtime=time(NULL);

    do {
      item=random()%i;
    } while ((genedpathname[item].whatisit!=ISFOLD));
    
    starttime+=time(NULL)-endtime;
    
    if ((genedpathname[i].pn=(char*)malloc(
	 strlen(genedpathname[item].pn)+9))==NULL) {
      printf("Out of memory\n");
      exit(7);
    }
      
    sprintf(genedpathname[i].pn, "%s:%X",genedpathname[item].pn,
	    random());
    
    genedpathname[i].whatisit=random()%3 < 2 ? ISFILE : ISFOLD;
    
    printf("#%i  %s - '%s'\n", i, genedpathname[i].whatisit==ISFOLD ? 
	   "FOLD" : "FILE", genedpathname[i].pn);
   
    if (genedpathname[i].whatisit==ISFOLD)  {
      if ((rc=maccreatefolder(&mr,genedpathname[i].pn))) {
	printf("maccreate returned %d\n", rc);
	macunmount(&mr);
	exit(3);
      }
    } else {
      if ((rc=maccreatefile(&mr,genedpathname[i].pn,1,&mfile))) {
	printf("maccreate returned %d\n", rc);
	macunmount(&mr);
	exit(3);
      }
      if ((rc=macclose(&mfile))) {
	printf("macclose returned %d\n", rc);
	macunmount(&mr);
	exit(4);
      }
    }

#ifdef CORRECTNESS

    if ((rc=checkvbmap(&mr))) {
      printf("checkvbmap returned %d\n", rc);
      macunmount(&mr);
      exit(9);
    }
    
    if ((rc=CheckFsBtree(&mr, &(mr.catalog), 1))) {
      printf("CheckFsBtree(catalog) Returned %d\n", rc);
      macunmount(&mr);
      exit(5);
    }
    
    if ((rc=CheckFsBtree(&mr, &(mr.extents), 0))) {
      printf("CheckFsBtree(extents) Returned %d\n", rc);
      macunmount(&mr);
      exit(6);
    }

#endif
    

  }

  endtime=time(NULL);
  printf("Time to generate %d files: %f seconds\n", MAXITEMS, 
	 difftime(endtime, starttime));

#ifdef CORRECTNESS
  puts("(Complete FS Checks Run After Each File)");
#endif


  printf("Now doing %d write-appends of %d bytes to %d files\n",
	 NUMWRITES, BLOCKSIZ, MAXITEMS);


  starttime=time(NULL);

  for (i=0;i<NUMWRITES;i++) {
    
    endtime=time(NULL);

    do {
      item=random()%MAXITEMS+1;
    } while (genedpathname[item].whatisit!=ISFILE);
    
    printf("#%d Trying to append to '%s'\n", i, genedpathname[item].pn);

    starttime+=time(NULL)-endtime;

    if ((rc=macopen(&mr, genedpathname[item].pn, WRITE | APPEND, &mfile))) {
      printf("macopen failed with rc=%d\n", rc);
      macunmount(&mr);
      exit(90);
    }

    curbs=random()%(BLOCKSIZ+1);

    printf("  Try to append %d bytes\n", curbs);

    if ((rc=macwrite(&mfile, curbs, buf, &actualbs))) {
      printf("macwrite failed with rc=%d\n", rc);
      macclose(&mfile);
      macunmount(&mr);
      exit(91);
    }
    
    if (actualbs!=curbs) {
      printf("WRITE FAILURE: Actually wrote %d bytes\n", actualbs);
    }

    if ((rc=macclose(&mfile))) {
      printf("macclose returned rc=%d\n", rc);
      macunmount(&mr);
      exit(92);
    }

#ifdef CORRECTNESS

    if ((rc=checkvbmap(&mr))) {
      printf("checkvbmap returned %d\n", rc);
      macunmount(&mr);
      exit(9);
    }
    
    if ((rc=CheckFsBtree(&mr, &(mr.catalog), 1))) {
      printf("CheckFsBtree(catalog) Returned %d\n", rc);
      macunmount(&mr);
      exit(5);
    }
    
    if ((rc=CheckFsBtree(&mr, &(mr.extents), 0))) {
      printf("CheckFsBtree(extents) Returned %d\n", rc);
      macunmount(&mr);
      exit(6);
    }
#endif

  }

  endtime=time(NULL);

  printf("Completed %d write-appends of MAX %d bytes each to %d files in %f seconds\n",
	 NUMWRITES, BLOCKSIZ, MAXITEMS, difftime(endtime, starttime));

  printf("Now Deleting Files and Folders in Random Order\n");

  i=1;

  starttime=time(NULL);
  
  while (i<=MAXITEMS) {

    endtime=time(NULL);

/* Random is not random enough */

    if (i<(MAXITEMS-MAXITEMS/5))
      do {
	item=random()%(MAXITEMS) + 1;
      } while ((genedpathname[item].whatisit==ISGONE));
    else 
      for (item=MAXITEMS; item>0 && (genedpathname[item].whatisit==ISGONE); 
	   item--);
    

    gentime=time(NULL);

/* Now we have a random item */
    
    printf("Deleting #%d  %s - '%s'\n", i, 
	   genedpathname[item].whatisit==ISFOLD ? 
	   "FOLD" : "FILE", genedpathname[item].pn);
   
    if (genedpathname[item].whatisit==ISFOLD)  {
      if ((rc=macdeletefolder(&mr,genedpathname[item].pn))) {
	if (rc==MACDELFOLDERNOTEMPTY) { 
	  printf("=>Can't Delete - Not Empty\n");
	  starttime+=time(NULL)-endtime;
	  continue;
	} else {
	  printf("macdeletefolder returned %d\n", rc);
	  macunmount(&mr);
	  exit(3);
	}
      }
    } else {
      if ((rc=macdeletefile(&mr,genedpathname[item].pn))) {
	printf("macdeletefile returned %d\n", rc);
	macunmount(&mr);
	exit(3);
      }
    }

    starttime+=(gentime-endtime);

    genedpathname[item].whatisit=ISGONE;

    ++i;

#ifdef CORRECTNESS

    if ((rc=checkvbmap(&mr))) {
      printf("checkvbmap returned %d\n", rc);
      macunmount(&mr);
      exit(9);
    }
    
    if ((rc=CheckFsBtree(&mr, &(mr.catalog), 1))) {
      printf("CheckFsBtree(catalog) Returned %d\n", rc);
      macunmount(&mr);
      exit(5);
    }
    
    if ((rc=CheckFsBtree(&mr, &(mr.extents), 0))) {
      printf("CheckFsBtree(extents) Returned %d\n", rc);
      macunmount(&mr);
      exit(6);
    }

#endif
    
  }

 
  macunmount(&mr);

  endtime=time(NULL);

  printf("Time to Delete %d files: %f seconds\n", MAXITEMS, 
	 difftime(endtime, starttime));
  puts("Done");

  exit(0);
 
}

    

    



