#include <time.h>
#include <stdio.h>
#include <fcntl.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <sys/time.h>

struct timeval tv;

void start_timing()
{
    gettimeofday(&tv, NULL);
}

double end_timing()
{
    struct timeval tv2;
    gettimeofday(&tv2, NULL);
    return(tv2.tv_sec - tv.tv_sec
	   + 1.e-6 * (tv2.tv_usec - tv.tv_usec));
}

#define NODEBUG

#define FILESIZE 5000000L
#define FILENAME "testfile"

int main()
{
  double TIME;
  long blocksize;
  int rc;
  int fd;
  
  
  printf("Running create/write-append/delete with varying blocksizes...\n");
#if 1
  for (blocksize=500L; blocksize<=50000L; blocksize*=10L) {
    printf("Writing %ld file with %ld size blocks...\n", FILESIZE, blocksize);
    start_timing();
    if ((fd=creat(FILENAME,0666))<0) {
      printf("Can't create file\n");
      exit(-1);
    }
    if ((rc=writefile(fd,blocksize))) {
      printf("write file return rc=%d\n", rc);
      exit(-1);
    }
    close(fd);	
    unlink(FILENAME);
    TIME=end_timing();
    printf("%lu byte file created/written/deleted in %f s or %f bytes/sec\n",
	   FILESIZE, TIME, ((float)FILESIZE)/ TIME);
  }
#endif
  printf("Creating file for over-writes\n");
  if ((fd=creat(FILENAME,0666))<0) {
    printf("Can't create file\n");
    exit(-1);
  }
  if ((rc=writefile(fd,5000L))) {
    printf("write file return rc=%d\n", rc);
    exit(-1);
  }
  close(fd);	

  printf("Overwriting file...\n");

  for (blocksize=500L; blocksize<=50000L; blocksize*=10L) {
    printf("Writing %ld file with %ld size blocks...\n", FILESIZE, blocksize);
    start_timing();
    if ((fd=open(FILENAME,O_WRONLY,0666))<0) {
      printf("Can't open file\n");
      exit(-1);
    }
    if ((rc=writefile(fd,blocksize))) {
      printf("write file return rc=%d\n", rc);
      exit(-1);
    }
    close(fd);	
    TIME = end_timing();
    printf("%lu byte file opened/written/closed in %f s or %f bytes/sec\n",
	   FILESIZE, TIME, ((float)FILESIZE)/ TIME);
  }

  printf("Reading file...\n");

  for (blocksize=500L; blocksize<=50000L; blocksize*=10L) {
    printf("Reading %ld file with %ld size blocks...\n", FILESIZE, blocksize);
    start_timing();
    if ((fd=open(FILENAME,O_RDONLY,0666))<0) {
      printf("Can't open file\n");
      exit(-1);
    }
    if ((rc=readfile(fd,blocksize))) {
      printf("write file return rc=%d\n", rc);
      exit(-1);
    }
    close(fd);	
    TIME = end_timing();
    printf("%lu byte file opened/read/closed in %s s or %f bytes/sec\n",
	   FILESIZE, TIME, ((float)FILESIZE)/ TIME);
  }

  unlink(FILENAME);

  
  
}




int writefile(int fd, long blocksize) 
{
  long writenum;	
  char *buf;
  unsigned rc;	
  
  
  if ((buf=(char*) malloc((size_t)blocksize))==NULL)
    return -1;	
  
  
  for (writenum=0;writenum<(FILESIZE/blocksize); writenum++) {
#ifdef DEBUG
    printf("Doing write number %ld - write(%d,buf,%u)\n", writenum, fd, (unsigned) blocksize);
#endif
    if ((rc=write(fd,buf,(unsigned)blocksize))!=blocksize) {
      printf("Write failed....\n");
      perror("write failed");
      free(buf);
      return rc;
    }
  }
  
  free(buf);
  
  return 0;
}




int readfile(int fd, long blocksize) 
{
  long readnum;	
  char *buf;
  unsigned rc;	
  
  
  if ((buf=(char*) malloc((size_t)blocksize))==NULL)
    return -1;	
  
  
  for (readnum=0;readnum<(FILESIZE/blocksize); readnum++) {
#ifdef DEBUG
    printf("Doing read number %ld - read(%d,buf,%u)\n", readnum, fd, (unsigned) blocksize);
#endif
    if ((rc=read(fd,buf,(unsigned)blocksize))!=blocksize) {
      printf("read failed....\n");
      free(buf);
      return rc;
    }
  }
  
  free(buf);
  
  return 0;
}



