#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <stdlib.h>
#include <sysent.h>
#include <errno.h>
#include <sys/stat.h>
#include <strings.h>
#include "../defs.h"
#include "../errors.h"
#include "../Lock/lock.h"
#include "../mountrec.h"
#include "../LibFuncs/macmount.h"
#include "../DevSupport/readwrite.h"
#include "../Volume/volinfo.h"
#include "../Volume/volbmap.h"
#include "../LibFuncs/catalog.h"
#include "../LibFuncs/catalog_local.h"
#include "../btree/btree.h"
#include "../LibFuncs/macreadwrite.h"
#include "../LibFuncs/catrectable.h"
#include "../Times/mac_time.h"

MOUNT_RECORD mr;

#define BLOCKSIZ 100000

void quit()
{
    macunmount(&mr);
    exit(0);
}


struct timeval tv;

void start_timing()
{
    gettimeofday(&tv, NULL);
}

double end_timing()
{
    struct timeval tv2;
    gettimeofday(&tv2, NULL);
    return(tv2.tv_sec - tv.tv_sec
	   + 1.e-6 * (tv2.tv_usec - tv.tv_usec));
}

int resolve_path(char *cdloc,char *dir,char *arg)
{
    struct stat sbuf;
    int is_file=1;
    FILE *f;
    
    if(stat(cdloc,&sbuf)== -1){
	if(ENOENT == errno){
	    is_file=0;
	} else {
	    printf("count not stat %s\n",cdloc);
	}
    }

    if(is_file && ((f=fopen(cdloc,"r+"))==NULL)){
	    printf("could not open : %s\n",cdloc);
	    quit();
    }

    strcpy(dir,arg);
    if (*(arg)!=':') {
	if(is_file){
	    fgets(dir,256,f);
	    strcat(dir,":");
	} else {
	    dir[0]=':';
	    dir[1]=NULL;
	}
    }
    strcat(dir,arg);
    if(is_file)fclose(f);
    return 0;
}

/* for create/del */
#define ITER_COUNT 1
#define NUM_FILES 1
/* for others */
#define IO_TIMES 2

int main(int argc, char *argv[])
{
    int blocksz, filesz, pass;
    int fail;
    MFILE mfile2;
    double TIME;
    int c2,c3;
    BYTE buf[BLOCKSIZ];
    LONGWORD byteswritten;
    char *devname, *partnum,*cdfileloc;
    BYTE dir[256],dir2[256],name[256];
    int cr;
    devname=getenv("MACDEVICE");
    partnum=getenv("MACPART");
    cdfileloc=getenv("MACCDLOC");

    
    if (devname==NULL || partnum==NULL || cdfileloc==NULL) {
	puts("Please set the environment variables MACDEVICE, MACPART and MACCDLOC before");
	puts("running any mac utility.  MACDEVICE is like /dev/rz4h or /dev/fd0");
	exit(1);
    }
    

    if (macmount(devname,atoi(partnum),&mr)) {
	printf("Sorry, couldn't mount %s partition %s\n", devname, partnum );
	exit(1);
    }
    
    initopencatrectable();
    resolve_path(cdfileloc,dir,"writetest");
    if ((cr=maccreatefile(&mr, dir, WRITE, &mfile2))) {
	printf("Can't Create %s %d\n",dir,cr);
	fail=1;
    }
    macclose(&mfile2);
		
    if (cr=macdeletefile(&mr, dir)){
	printf("Can't delete %s %d\n",dir,cr);
	fail=1;
    }


    resolve_path(cdfileloc,dir2,"");
    start_timing();
    for(pass=0;pass<ITER_COUNT;pass++)
    {
	int i;
	for(i=0;i<NUM_FILES;i++) {
	    sprintf(name,"%stest%d",dir2,i);
	
	    if ((cr=maccreatefile(&mr, name, WRITE, &mfile2))) {
		printf("Can't Create %s %d\n",name,cr);
		quit();
	    }
	    macclose(&mfile2);
	}
	for(i=0;i<NUM_FILES;i++)
	    sprintf(name,"%stest%d",dir2,i);
	if (macdeletefile(&mr, name)){
	    printf("Can't delete %s\n",name);
	    quit();
	}
    }
    TIME = end_timing();
    printf("time to create and delete %d files = %f sec = %f files/sec\n",
	   NUM_FILES*ITER_COUNT, TIME, (NUM_FILES*ITER_COUNT)/TIME);

    filesz = 5000 * 1000;
    for(blocksz=500;blocksz<=50000;blocksz*=10){
	start_timing();
	for(pass=0;pass<IO_TIMES;pass++){
	    if ((cr=maccreatefile(&mr, dir, WRITE, &mfile2))) {
		printf("Can't Create %s%d\n",dir,cr);
		quit();
	    }
	    c3=filesz/blocksz;
	    for(c2=0;c2<c3;c2++){
		if ((cr=macwrite(&mfile2,blocksz, buf, &byteswritten))){
		    printf("macwrite failed %d\n",cr);
		    quit();
		}
	    }
	    macclose(&mfile2);

	    if(pass != IO_TIMES-1)
		if ((cr=macdeletefile(&mr,dir))){
		    printf("Can't delete %s %d\n",dir,cr);
		    quit();
		}
	}
	TIME = end_timing();
	printf("time to write %d files (block size %08d) = %f = %011f bytes/sec\n",
	       IO_TIMES, blocksz, TIME,(filesz*IO_TIMES)/(double)(TIME));

	if ((cr=macopen(&mr, dir, READ | WRITE, &mfile2))) {
	    printf("Can't open %s %d\n",dir,cr);
	    quit();
	}
	start_timing();
	for(pass=0;pass<IO_TIMES;pass++){
	    c3=filesz/blocksz;
	    for(c2=0;c2<c3;c2++){
		if ((cr=macwrite(&mfile2,blocksz, buf, &byteswritten))){
		    printf("macwrite failed %d\n",cr);
		    quit();
		}
	    }
	    macseek(&mfile2, 0, SEEKFROMBEGINING);
	}
	TIME = end_timing();
	printf("time to overwrite %d times (blocksz %08d) = %f = %011f bytes/sec\n",
	       IO_TIMES, blocksz, TIME,(filesz*IO_TIMES)/(TIME));
	start_timing();
	for(pass=0;pass<IO_TIMES;pass++)
	{
	    macseek(&mfile2, 0, SEEKFROMBEGINING);
	    c3=filesz/blocksz;
	    for(c2=0;c2<c3;c2++){
		if ((cr=macread(&mfile2,blocksz, buf, &byteswritten))){
		    printf("macread failed %d\n",cr);
		    quit();
		}
	    }
		
	}
	TIME = end_timing();
	printf("time to read %d files (blocksz = %08d) = %f = %011f bytes/sec\n",
	       IO_TIMES,blocksz,TIME,(IO_TIMES*filesz)/(TIME));
	    
	macclose(&mfile2);
	if ((cr=macdeletefile(&mr,dir))){
	    printf("Can't delete %s %d\n",dir,cr);
	    quit();
	}
    }
    macunmount(&mr);
    exit(0); 
}
