#include <stdio.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/uio.h>
#include <sys/file.h>
#include <stdlib.h>
#include <sysent.h>
#include <errno.h>
#include <sys/stat.h>
#include <strings.h>
#include "../defs.h"
#include "../errors.h"
#include "../Lock/lock.h"
#include "../mountrec.h"
#include "../LibFuncs/macmount.h"
#include "../DevSupport/readwrite.h"
#include "../Volume/volinfo.h"
#include "../Volume/volbmap.h"
#include "../LibFuncs/catalog.h"
#include "../LibFuncs/catalog_local.h"
#include "../btree/btree.h"
#include "../LibFuncs/macreadwrite.h"
#include "../LibFuncs/catrectable.h"
#include "../Times/mac_time.h"

MOUNT_RECORD mr;

#define BLOCKSIZ 1000

void quit()
{
    macunmount(&mr);
    exit(0);
}


int resolve_path(char *cdloc,char *dir,char *arg)
{
    struct stat sbuf;
    int is_file=1;
    FILE *f;
    
    if(stat(cdloc,&sbuf)== -1){
	if(ENOENT == errno){
	    is_file=0;
	} else {
	    printf("count not stat %s\n",cdloc);
	}
    }

    if(is_file && ((f=fopen(cdloc,"r+"))==NULL)){
	    printf("could not open : %s\n",cdloc);
	    quit();
    }

    strcpy(dir,arg);
    if (*(arg)!=':') {
	if(is_file){
	    fgets(dir,256,f);
	    strcat(dir,":");
	} else {
	    dir[0]=':';
	    dir[1]=NULL;
	}
    }
    strcat(dir,arg);
    if(is_file)fclose(f);
    return 0;
}

int main(int argc, char *argv[])
{
    int fail;
    MFILE mfile1,mfile2;
    int b;
    int i,j;
    CATALOG_RECORD CR;
    BYTE buf[BLOCKSIZ];
    LONGWORD bytesread, byteswritten;
    char *devname, *partnum,*cdfileloc,*ind;
    BYTE dir[256],dir2[256],name[256];
    
    devname=getenv("MACDEVICE");
    partnum=getenv("MACPART");
    cdfileloc=getenv("MACCDLOC");

    
    if (devname==NULL || partnum==NULL || cdfileloc==NULL) {
	puts("Please set the environment variables MACDEVICE, MACPART and MACCDLOC before");
	puts("running any mac utility.  MACDEVICE is like /dev/rz4h or /dev/fd0");
	exit(1);
    }
    
    
    if (argc!=3){
	puts("usage: maccp :my:mac:file :my:mac:file");
	exit(0);
    }
    
    if (macmount(devname,atoi(partnum),&mr)) {
	printf("Sorry, couldn't mount %s partition %s\n", devname, partnum );
	exit(1);
    }
    
    initopencatrectable();
    
    fail=0;
    
    b=0;

    resolve_path(cdfileloc,dir,argv[1]);
    resolve_path(cdfileloc,dir2,argv[2]);

    
    
    ind = rindex(argv[0],'/');
    if(strcmp( ((ind)?ind:argv[0]),"macmv") == 0){
	if(macmove(&mr,dir,dir2)){
	    puts("macmove failed\n");
	}
    } else {
	
	if ((maccreatefile(&mr, dir2, WRITE, &mfile2))) {
	    printf("Can't Create %s\n", dir2);
	    fail=1;
	}
	if ((macopen(&mr, dir, READ, &mfile1))) {
	    printf("Can't Open %s\n",dir);
	    fail=1;
	}
	if (!fail) {
	    do {
		
		b++;
		if (macread(&mfile1, BLOCKSIZ, buf, &bytesread)) {
		    puts("macread failed");
		    break;
		}
		if (macwrite(&mfile2, bytesread, buf, &byteswritten)){
		    puts("macwrite failed");
		    break;
		}
		/*
		  printf("vbmap:%d\n", checkvbmap(&mr));
		  */
	    } while (bytesread==BLOCKSIZ && bytesread==byteswritten);
	    macclose(&mfile1);
	    macclose(&mfile2);
	}
    }
    
    macunmount(&mr);
    
    exit(0); 
    
}
