#include <stdio.h>
					/* Test program for BTREE routines */

#include "../defs.h"
#include "../mountrec.h"
#include "../DevSupport/readwrite.h"
#include "local_btree.h"
#include "btree.h"
#include "../errors.h"
#include "../LibFuncs/macmount.h"

int srandom(int);
int random(void);

MOUNT_RECORD mnt;
EXTENT_RECORD ext;
CATALOG_RECORD cat;
char inserted_files[3000][15];

int Loops    = 100;
int Inserts  = 0;
int Deletes  = 0;
int Searches = 0;

void main()    {

  int err;
  LONGWORD foundid;
  char foundname[32];
  int j;
  int inserted;
			      /* Init */
  err = macmount("/afs/cs/user/pdinda/OSProj/DiskImages/Simple.test", 
		 3, & mnt);
  if(err)
    goto out;
    
  err = CheckFsBtree(&mnt, & mnt.catalog, 1); 
  if(err)
    goto out;
  
			      /* Delete everything from catalog in Simple 
                               * after that starts inserting random files 
                               * until max 1000 files or until 
                               * BT_UNIMPLEMENTED is returned. After that 
                               * starts to cancel the inserted files in the 
                               * order in which were inserted */
					

			      /* Save for future use a catalog entry */
  Searches++;
  err = GetCatalogRecord(&mnt, 2, "test1", &cat, &foundid, foundname);
  if(err)
    goto out;
  PutString("testprogram", cat.U.cdrThdRec.thdCName);
  
					
			      /* Starts deleting */
			      /* Desktop */
  Deletes++;
  err = DelCatalogRecord(&mnt, 2, "Desktop");
  if(err)
    goto out;

  err = CheckFsBtree(&mnt, & mnt.catalog, 1); 
  if(err)
    goto out;
			      /* Dir thread record */
  Deletes++;
  err = DelCatalogRecord(&mnt, 2, "");
  if(err)
    goto out;

  err = CheckFsBtree(&mnt, & mnt.catalog, 1); 
  if(err)
    goto out; 

			      /* Parent of root  directory  */
  Deletes++;
  err = DelCatalogRecord(&mnt, 1, "Simple Test Disk");
  if(err)
    goto out;

  err = CheckFsBtree(&mnt, & mnt.catalog, 1); 
  if(err)
    goto out; 

  
			      /* Test1 */
  Deletes++;
  err = DelCatalogRecord(&mnt, 2, "test1");
  if(err)
    goto out;

  err = CheckFsBtree(&mnt, & mnt.catalog, 1); 
  if(err)
    goto out;

			      /* Now the catalog should be empty */
  if(mnt.catalog.bth.bthRoot != 0) {
    err = 1000;
    goto out;
  }
					
Insert:

  printf("\nNow inserting...\n");					
					
			      /* Starts inserting */
					
  for(inserted=0; inserted < 1000;inserted++)  {
    cat.cdrType = random()%3 + 1;

    Inserts++;

    sprintf(inserted_files[inserted], "test%x", random());
    printf("Loop %d -- Inserting %d -- %s", Loops, Inserts,
	   inserted_files[inserted]);


    err = PutCatalogRecord(&mnt, 2, inserted_files[inserted], &cat);
    if(err)  {
      if(err == BT_UNIMPLEMENTED) {
	err = 0;
	break;
      }
      goto out;
    }
			      /* INSERTED is the number of files that were 
                               * inserted  */

    err = CheckFsBtree(&mnt, & mnt.catalog, 1);
    if(err)
      goto out; 
    printf(" --- Ok\n");
    
  }

  
  printf("\nNow searching...\n");

  for(j=0; j< inserted; j++)   {
    Searches++;
    printf("Loop %d -- Searching %d -- %s", Loops, Searches,
	   inserted_files[j]);

    err = GetCatalogRecord(&mnt, 2, inserted_files[j], &cat, &foundid,
			   foundname);
    if(err)
      goto out;
    printf(" --- OK\n");
  }


  err = CheckFsBtree(&mnt, & mnt.catalog, 1);
  if(err)
    goto out; 


    
  printf("\nNow deleting...\n");
					
			      /* Delete all the inserted files */
  for(j=0;j<inserted;j++) {
    Deletes++;
    printf("Loop %d -- Deleting %d -- %s", Loops, Deletes,
	   inserted_files[j]);
    err = DelCatalogRecord(&mnt, 2, inserted_files[j]);
    if(err)
      goto out;

    err = CheckFsBtree(&mnt, & mnt.catalog, 1);
    if(err) 
      goto out; 
    printf(" --- Ok\n");
  }

			      /* Now the tree must be empty */
  if(mnt.catalog.bth.bthRoot != 0) {
    err = 1000;
    goto out;
  }

  if(--Loops > 0)
    goto Insert;					
					
 out:
  printf("\nExit because err = %d\n", err);
  err = macunmount(& mnt);
  if(err)
    printf("Unmount Failed!");

  exit(0);
}




