/* Internal declarations for macfs */

#define VFS
#include "macfs.h"
#include "../Includes/macvfslevel.h"

#define MAC_AHEAD_BY 0x7C25783F
#define TIME_MACTOUX(A) ((A) - MAC_AHEAD_BY)
#define TIME_UXTOMAC(A) ((A) + MAC_AHEAD_BY)
extern struct vnode *specvp(), *dnlc_lookup();
extern int einval();
#define ISVDEV(t) (((t) == VCHR) || ((t) == VBLK))

/* Constants: TRUE FALSE MAXFNLENGTH BYTESPERSEC
              ROOTPID ROOTID SMALLESTFN
   Types: BYTE WORD LONGWORD SIGNEDLONGWORD
          MOUNT_RECORD CATALOG_RECORD CATRECANDNAME
   Functions:
     macmount macunmount syncmountrecord

     mdirentries mopen mclose openroot
     mcreate{file,folder} mmove{file,folder} mdelete{file,folder}

     mread mwrite mgrowfile mtruncatefile

   MOUNT_RECORD: isreadable() iswriteable() isappend() isshared()
     vib.CrDate, vib.LsMod,
     vib.NmAlBlks, vib.AlBlkSize, vib.FreeBks, vib.ClpSiz
     vib.FilCnt, vib.DirCnt
   CATALOG_RECORD:
     cdrType = CDR_DIR, CDR_FILE, CDR_DTHREAD, CDR_FTHREAD
       U.cdrDirRec:
         dirDirID, dirCrDat, dirMdDat, dirVal
       U.cdrFilRec:
         filFlNum, filCrDat, filMdDat, filLgLen
       U.cdrThdRec:
         fthdParID fthdCName
     status = DELETED, MOVED
   CATRECANDNAME:
     cr
     name
*/

/* if set, i_mnt->im_mode = args.mode, else the modes are even for all */
#define MACFSMNT_PROT (1<<0)
struct macfs_args
{
    char *fspec;
    unsigned int flags;
    int partition;
    uid_t uid;
    gid_t gid;
    int mode;
};

struct macfs_mount
{
    struct vfs *vfsp;
    MOUNT_RECORD mrec;        /* to call the Unix innocent routines */
    struct vnode *im_devvp;   /* the covered vnode */
    int	im_flags;             /* MACFSMNT_TRANS, MACFSMNT_MUSER */
    dev_t	im_dev;       /* the device we do IO on (?) */
    uid_t	im_uid;       /* owner's user id */
    gid_t	im_gid;       /* owner's group id */
    int	im_mode;              /* mode protection bits */
};
#define VFSTOMACFS(_vfs_) ((struct macfs_mount *)(_vfs_)->vfs_data)

struct macfs_node
{
	struct	vnode i_vnode;	/* vnode associated with this inode */
	struct	macfs_mount *i_mnt; /* filesystem associated with this inode */
	CATALOG_RECORD crec; /* describes this fileid */
	/* protection records */
	uid_t	o_uid;    /* owner's user id */
	gid_t	o_gid;    /* owner's group id */
	/* Mode protection bits:
	   vnode->v_mode = (o_mode & VFMT) = VDIR or VREG (or VLNK);

	   IFDIR and IFREG are synonyms for VDIR and VREG?
	   VREAD and VWRITE are synonyms for IREAD and IWRITE?

	   own_bits = o_mode
	   group_bits = o_mode >> 3
	   world_bits = o_mode >> 6

	   permissions = (*_bits) & (IREAD, IWRITE, IEXEC)

	   Also have bits ISUID and ISGID for executables.
	   Also have bits dealing with special files.

	   We have no executables or special files or symbolic links, so
	   ip->o_mode = i_vnode->v_mode | i_mnt->im_mode
	*/
	u_short	o_mode;
	ino_t	o_number; /* file number or directory number */
	short	o_nlink;  /* # links -- 1 normal, 0 means should be deld */

	dev_t	o_dev;	  /* device we do IO on (?) */
	LONGWORD ParentId;
	char name[MAXFNLENGTH+1]; /* for self-deleting on release */
};
#define VTO_MAC(_vp_) ((struct macfs_node *)(_vp_))
#define MAC_TOV(_ip_) ((struct vnode *)(_ip_))

/* Name lookup caching and vnode integrity -- <= 1 vnode structure per
   file.

Now this is tricky.  We need to have <= 1 vnode structure per file
id, but we need to address our cache by (ParentId, name).

So, we use vnode = dnlc_lookup(vnode, name, cred),
                   dnlc_enter(vnode, name, child, NOCRED)
                   dnlc_remove(vnode, name)
                   dnlc_purge_vp(vnode)

But sometimes we can't because we have not parent node sometimes.  Oh well
for now do without dlnc_* entirely.

When we get a catalog record with mopen (or whatever), we need to make
sure that noone else has a vnode for it yet.  We do this by hashing on
(dev, o_number), and the crec_to_macfsnode call will lock the
hashtable, if the macfsnode exists then return it, else create the
macfsnode, add it to the table, and return that.
*/

#define	INOHSZ	128
#define INOHASH(dev, ino) ( ((int)(dev) + (int)(ino)) % (INOHSZ-1))

union macfs_ihead {
	union  macfs_ihead *ih_head[2];
	struct macfs_node *ih_chain[2];
} macfs_ihead[INOHSZ];

extern zone_t macfs_zone;

/* Written. */
int macfs_mountfs(
	struct vnode **devvpp,
	char		*path,
	struct vfs	*vfsp,
	struct macfs_args *args);

/* gibberish to me.  Calls macfs_mountfs. Written. */
int macfs_mount(
	struct vfs *vfsp,
	char *path,
	caddr_t data);

/* gibberish to me.  Calls macfs_mountfs. Written. */
int macfs_mountroot(void);

/* sync, iflush, gibberish,and macunmount. Written. */
int macfs_unmount(struct vfs *);

/* return root vnode of fs.  Written. */
int macfs_root(struct vfs *, struct vnode **);

/* write states from the mount record. Written */
int macfs_statfs(struct vfs *, struct statfs *);

/* calls syncmountrecord. Written. */
int macfs_sync(struct vfs *);

/* HOLD -- Written */
int macfs_open(struct vnode **, int flag, struct ucred *);

/* RELE -- Written */
int macfs_close(struct vnode *, int flag, struct ucred *);

/* hacks offsets, calls mread/mwrite, Written. */
int macfs_rdwr(
	struct vnode	*vp,
	register struct uio *uio,
	enum uio_rw	rw,
	int		ioflag,
	struct ucred	*cred);

/* return einval. Written. */
int macfs_ioctl(
	struct vnode	*vp,
	unsigned int	com,
	caddr_t		data,
	int		fflag,
	struct ucred	*cred);

/* return einval. Written. */
int macfs_select(
	struct vnode	*vp,
	int		which,
	struct ucred	*cred);

/* writes info from mrec and crec. Written. */
int macfs_getattr(
	struct vnode	*vp,
	register struct vattr *vap,
	struct ucred	*cred);

/* Calls macfs_chown1, macfs_itrunc do most of the work.
   Uses macfs_iupdat to mark the change. Written. */
int macfs_setattr(
register struct vnode *vp,
	register struct vattr *vap,
	struct ucred *cred);

/* calls macfs_iaccess to do the work. Written. */
int macfs_access(
	struct vnode	*vp,
	register int	mode,
	struct ucred	*cred);

/* Calls macfs_dirlook for most of work.  Written. */
int macfs_lookup(
	struct vnode	*vdp,
	char		*nm,
	struct vnode	**vpp,
	struct ucred	*cred);

/* won't create a directory.  Calls macfs_direnter(DE_CREATE). Then
plays with exclusive flags, checks macfs_iaccess permissions.  Uses
macfs_iput or macfs_itrunc.  Has gibberish for special vnodes.
underimplemented.
*/
int macfs_create(
	struct vnode	*dvp,
	char		*nm,
	struct vattr	*vap,
	enum vcexcl	exclusive,
	int		mode,
	struct vnode	**vpp,
	struct ucred	*cred);

/* calls macfs_dirremove. underimplemented */
int macfs_remove(
	struct vnode	*vp,
	char		*nm,
	struct ucred	*cred);

/* einval:
macfs_link();
*/

/* checks write permission on source directory,
   calls dirlook to find the source vnode,
   prevents rename of . or .. or alias of .
   calls direnter(DE_RENAME)
   removes the source entry (dirremove),
   calls macfs_irelease
underimplemented
*/
int macfs_rename(
	struct vnode	*sdvp,		/* old (source) parent vnode */
	char		*snm,		/* old (source) entry name */
	struct vnode	*tdvp,		/* new (target) parent vnode */
	char		*tnm,		/* new (target) entry name */
	struct ucred	*cred);

/* calls direnter(DE_CREATE). underimplemented */
int macfs_mkdir(
	struct vnode	*dvp,
	char		*nm,
	register struct vattr *vap,
	struct vnode	**vpp,
	struct ucred	*cred);


/* calls dirremove. underimplemented */
int macfs_rmdir(
	struct vnode	*vp,
	char		*nm,
	struct ucred	*cred);

/* passes back part of info from mdirentries.  Written.
*/
int macfs_readdir(
	struct vnode	*vp,
	register struct uio *uio,
	struct ucred	*cred);

/* returns einval. Written. */
int macfs_symlink(
	struct vnode	*dvp,
	char		*lnm,
	struct vattr	*vap,
	char		*tnm,
	struct ucred	*cred);

/* returns einval. Written. */
int macfs_readlink(
	struct vnode	*vp,
	struct uio	*uiop,
	struct ucred	*cred);

/* flush catalog changes, free node.  Written */
int macfs_inactive(struct vnode *, struct ucred *);

/* returns the device's strategy. Written */
int macfs_strategy(
	register struct buf *bp);

/* Unimplement or */
int macfs_zfree(struct macfs_node *);

/* einval
macfs_lockctl();
 */

/* make fid. hideous. underimplemented. */
int macfs_fid(
	struct vnode *vp,
	struct fid **fidpp);

/* free fid. hideous. underimplemented */
int macfs_freefid(
	struct vnode *vp,
	struct fid *fidp);

/* file handle to vnode. hideous. underimplemented */
int macfs_fhtovp(
	struct vfs *vfsp,
	struct vnode	**vpp,
	struct fid	*fhp);

/* simple. Written. */
int macfs_nlinks(struct vnode *vp, int *l, struct ucred *cred);

/* calls iupdat. Written */
int macfs_vupdate(
	struct vnode	*vp,
	int		flag);

/* tries to rm access to dev.  uses inode_uncache_try. Written */
int macfs_iflush(dev_t dev);

/* check access rights on node. Written. */
int macfs_iaccess(
	struct macfs_node *ip,
	int		  mode);

/* just release vnode part, checking that we've released the lock. 
   Written. */
void macfs_irelease(struct macfs_node *ip);

/* set file size.  underimplemented */
int macfs_itrunc(struct macfs_node *ip, int size);

/* notify change to node.  Calls mclose() to flush
   the crec to disk (if writable fs). Ignores waitfor. Written */
void macfs_iupdat(
	struct macfs_node	*ip,
	int			waitfor);
    
/* simple mode hacking and permission checking. Written. */
int macfs_chown1(struct macfs_node *ip, uid_t uid, gid_t gid);

/* Written */
void macfs_ilock(struct macfs_node *);

/* Written */
void macfs_iunlock(struct macfs_node *);

/*    (parent macfs_node, name, &macfs_node, cred) Written.
   Uses finddir, getnode, and crec_to_macfs_node to do work. */
int macfs_dirlook(struct macfs_node *, char *,
		  struct vnode **, struct ucred *);

/* Look in cache or return NULL.  The node will be locked and held.
Written but without dnlc_lookup -- just returns NULL. */
struct macfs_node *macfs_getnode(struct macfs_mount *,
                                 LONGWORD ParentId,
                                 char *name);

/* Find directory from its id -- works by checking for root
   or using SMALLESTFN to get the directory thread record...
   Returned node is locked and referenced.  Written. */
struct macfs_node *macfs_finddir(struct macfs_mount *i_mnt, LONGWORD did);

/* This should be called until it returns 0, at which point *mp will
   be a locked and held node or NULL */
int macfs_dev_and_num_lookup(dev_t dev, int o_number, struct macfs_node **mp);

/* Return new locked and held node for catalog record described.  Will
prevent multiple vnode/file problems by checking the hashlist for
the corresponding (dev,fid) pair before creating a vnode. Written. */
struct macfs_node *crec_to_macfsnode(struct macfs_mount *,
                                     CATALOG_RECORD *,
                                     int ParentId,
                                     char *name);

#if 0
/*
    when we get to direnter from rename:
      we can write to the source directory
      we have the macfs_vnode for the source
      the source entry != source directory
      we should probably pass source name in to direnter
    when we get to direnter from create:
      source isn't a directory
    when we get to direnter from mkdir
      nothing
    if name is . or .. fail
    if name is illlegal fail
    if renaming and source deleted under us fail
    lock target directory
    if target directory not a directory fail
    if target directory deleted under us fail
    if no execute access for target directroy fail
    if renaming a directory to a new parent and parent not writable fail
    if renaming a directory to a new parent and cycling fail
    if search for target ip errors then fail
    if target exists
       if creating (if ipp=NULL set error = EEXIST else put the target)
       if renaming (if source=target error ESAME else call dirrename)
                   put the target
    else
       if cannot write target directory fail
       if renaming and sourcedir sticky and sourcedir != targetdir and
          not owner of sourcedir or sourcefile then error EPERM
       if creating, make new inode at sip
       add directory entry sip into tdp
       if failure
          if creating, release the new node
       else if saving source, save it
       else if creating release new node
    unlock the target directory
*/
direnter

dirremove

dirbadname

dirmangled

dirbad

dircheckpath

dircheckforname

dirrename

diraddentry

dirempty

macfs_iget

macfs_ialloc

macfs_iput

macfs_zfree

macfs_iflush

macfs_inactive

macfs_itdiren
#endif
