/* BOOGER */
int macfs_create(
	struct vnode	*dvp,
	char		*nm,
	struct vattr	*vap,
	enum vcexcl	exclusive,
	int		mode,
	struct vnode	**vpp,
	struct ucred	*cred)
{
    register int error;
    struct macfs_node *ip;

trace(macfs_debug,("\nDcreate "));
    /*
     * can't create directories. use macfs_mkdir.
     */
    if ((vap->va_mode&IFMT) == VDIR)
	return (EISDIR);
    ip = (struct macfs_node *) 0;
    error = macfs_direnter(VTO_MAC(dvp), nm, DE_CREATE,
			   (struct macfs_node *)0, (struct macfs_node *)0,
			   vap, &ip, cred);
    /*
     * if file exists and this is a nonexclusive create,
     * check for not directory and access permissions
     * If create/read-only an existing directory, allow it.
     */
    if (error == EEXIST) {
	if (exclusive == NONEXCL) {
	    if (((ip->o_mode & IFMT) == IFDIR) && (mode & IWRITE)) {
		error = EISDIR;
	    } else if (mode) {
		error = macfs_iaccess(ip, mode);
	    } else {
		error = 0;
	    }
	}
	if (error) {
	    macfs_iput(ip);
	} else if (((ip->o_mode&IFMT) == IFREG) && (vap->va_size == 0)){
	    /*
	     * truncate regular files, if required
	     */
	    macfs_itrunc(ip, (u_long) 0);
	}
    } 
    if (error) {
	return (error);
    }
    *vpp = MAC_TOV(ip);
    MAC_IUNLOCK(ip);
    /*
     * If vnode is a device return special vnode instead
     */
    if (ISVDEV(((*vpp)->v_mode&VFMT))) {
	struct vnode *newvp, *specvp();
	
	newvp = specvp(*vpp, (*vpp)->v_rdev);
	VN_RELE(*vpp);
	*vpp = newvp;
    }
    
    if (vap != (struct vattr *)0) {
	(void) VOP_GETATTR(*vpp, vap, cred);
    }
    return (error);
}

int macfs_remove(
	struct vnode	*vp,
	char		*nm,
	struct ucred	*cred)
{
trace(macfs_debug,("\nDremove "));
	return macfs_dirremove(VTO_MAC(vp), nm, (struct macfs_node *)0, 0);
}


/*
 * Rename a file or directory
 * We are given the vnode and entry string of the source and the
 * vnode and entry string of the place we want to move the source to
 * (the target). The essential operation is:
 *	unlink(target);
 *	link(source, target);
 *	unlink(source);
 * but "atomically".
 * Note that we do not have hard links, so things are easier than UFS.
 */
/* BOOGER */
int macfs_rename(
	struct vnode	*sdvp,		/* old (source) parent vnode */
	char		*snm,		/* old (source) entry name */
	struct vnode	*tdvp,		/* new (target) parent vnode */
	char		*tnm,		/* new (target) entry name */
	struct ucred	*cred)
{
	struct macfs_node *sip;		/* source inode */
	register struct macfs_node *sdp;	/* old (source) parent inode */
	register struct macfs_node *tdp;	/* new (target) parent inode */
	register int error;

trace(macfs_debug,("\nDrename "));
	sdp = VTO_MAC(sdvp);
	tdp = VTO_MAC(tdvp);
	/*
	 * make sure we can delete the source entry
	 */
	error = macfs_iaccess(sdp, IWRITE);
	if (error) {
		return (error);
	}
	/*
	 * look up inode of file we're supposed to rename.
	 */
	error = macfs_dirlook(sdp, snm, (struct vnode **)&sip, cred);
	if (error) {
		return (error);
	}

	MAC_IUNLOCK(sip);			/* unlock inode (it's held) */
	/*
	 * check for renaming '.' or '..' or alias of '.'
	 */
	if ((strcmp(snm, ".") == 0) || (strcmp(snm, "..") == 0) ||
	    (sdp == sip)) {
		error = EINVAL;
		goto out;
	}
	/*
	 * link source to the target
	 */
	error =
	    macfs_direnter(tdp, tnm, DE_RENAME,
		sdp, sip, (struct vattr *)0, (struct macfs_node **)0, cred);
	if (error)
		goto out;

	/*
	 * Unlink the source
	 * Remove the source entry. Dirremove checks that the entry
	 * still reflects sip, and returns an error if it doesn't.
	 * If the entry has changed just forget about it. 
	 * Release the source inode.
	 */
	error = macfs_dirremove(sdp, snm, sip, 0);
	if (error == ENOENT)
		error = 0;
out:
	macfs_irelease(sip);
	return (error);
}

int macfs_mkdir(
	struct vnode	*dvp,
	char		*nm,
	register struct vattr *vap,
	struct vnode	**vpp,
	struct ucred	*cred)
{
	struct macfs_node *ip;
	register int error;

trace(macfs_debug,("\nDmkdir "));
	error =
	    macfs_direnter(VTO_MAC(dvp), nm, DE_CREATE,
		(struct macfs_node *)0, (struct macfs_node *)0, vap, &ip, cred);
	if (error == 0) {
		*vpp = MAC_TOV(ip);
		MAC_IUNLOCK(ip);
	} else if (error == EEXIST) {
		macfs_iput(ip);
	}
	return (error);
}

int macfs_rmdir(
	struct vnode	*vp,
	char		*nm,
	struct ucred	*cred)
{
trace(macfs_debug,("\nDrmdir "));
	return macfs_dirremove(VTO_MAC(vp), nm, (struct macfs_node *)0, 1);
}

/*
 * Write a new directory entry.
 * The directory must not have been removed and must be writeable.
 * There are three operations in building the new entry: creating a file
 * or directory (DE_CREATE), renaming (DE_RENAME) or linking (DE_LINK).
 * Since MAC does not have hard links only the first two apply.
 * There are four possible cases to consider:
 *	Name
 *	found	op			action
 *	----	---			-------------------------------
 *	no	DE_CREATE		create file according to vap and enter
 *	no	DE_RENAME		enter the file sip
 *	yes	DE_CREATE		error EEXIST *ipp = found file
 *	yes	DE_RENAME		remove existing file, enter new file
 */
int macfs_direnter(
	register struct macfs_node *tdp,/* target directory to make entry in */
	register char		*namep,	/* name of entry */
	enum de_op 		op,	/* entry operation */
	register struct macfs_node *sdp,/* source inode parent if rename */
	struct macfs_node	*sip,	/* source inode if link/rename */
	struct vattr		*vap,	/* attributes if new inode needed */
	struct macfs_node	**ipp,	/* return entered inode (locked) here */
	struct ucred		*cred)	/* credentials to use */
{
	struct macfs_node *tip;		/* inode of (existing) target file */
#if 0
	struct slot slot;		/* slot info to pass around */
#endif
	register int namlen;		/* length of name */
	register int error;		/* error number */
	struct macfs_mount *imp;	/* used to get translation type */

trace(macfs_debug,("Ddirent %x %s %x ", tdp->o_number, namep, op));

	if (ipp)
		*ipp = NULL;

	if (op == DE_LINK)
		return EINVAL;

	namlen = strlen(namep);
	/*
	 * If name is "." or ".." then if this is a create look it up
	 * and return EEXIST. Rename or link TO "." or ".." is forbidden.
	 */
	if (namlen == 1 && namep[0] == '.' ||
	    namlen == 2 && namep[0] == '.' && namep[1] == '.') {
		if (op == DE_RENAME) {
			return (ENOTEMPTY);
		}
		if (ipp) {
			if (error = macfs_dirlook(tdp, namep,
					(struct vnode **) ipp, cred))
				return (error);
		}
		return (EEXIST);
	}

	imp = tdp->i_mnt;
#if 0
	if (macfs_dirbadname(namep, namlen, (imp->im_flags & MACFSMNT_TRANS)))
		return (EPERM);

	slot.status = NONE;
	slot.bp = NULL;
#endif
	/*
	 * For rename lock the source entry and check the link count to
	 * see if it has been removed while it was unlocked.
	 */
	if (op != DE_CREATE) {
		MAC_ILOCK(sip);
		if (sip->o_nlink == 0) {
			MAC_IUNLOCK(sip);
			return (ENOENT);
		}
		MAC_IUNLOCK(sip);
	}
	/*
	 * lock the directory in which we are trying to make the new entry.
	 */
	MAC_ILOCK(tdp);
	/*
	 * Check accessiblity of directory.
	 */
	if ((tdp->o_mode&IFMT) != IFDIR) {
		error = ENOTDIR;
		goto out;
	}
	/*
	 * If target directory has not been removed, then we can consider
	 * allowing file to be created.
	 */
	if (tdp->o_nlink == 0) {
		error = ENOENT;
		goto out;
	}
	/*
	 * Execute access is required to search the directory.
	 */
	if (error = macfs_iaccess(tdp, IEXEC)) {
		goto out;
	}
	/*
	 * If this is a rename and we are doing a directory and the parent
	 * is different (".." must be changed), then the source directory must
	 * not be in the directory heirarchy above the target, as this would
	 * orphan everything below the source directory. Also the user must
	 * have write permission in the source so as to be able to change "..".
	 */
	if ((op == DE_RENAME) &&
	    ((sip->o_mode&IFMT) == IFDIR) && (sdp != tdp)) {
		if (error = macfs_iaccess(sip, IWRITE))
			goto out;
		if (error = macfs_dircheckpath(sip, tdp))
			goto out;
	}
	/*
	 * search for the entry
	 */
	error = macfs_dircheckforname(tdp, namep, &tip);
	if (error) {
		goto out;
	}

	if (tip) {
		switch (op) {

		case DE_CREATE:
			if (ipp) {
				*ipp = tip;
				error = EEXIST;
			} else {
				macfs_iput(tip); /* BOOGER? no error */
			}
			break;

		case DE_RENAME:
			if (sip->o_number == tip->o_number) {
				/*
				 * Short circuit rename (foo, foo).
				 */
				error = ESAME;
			} else {
				error =
				    macfs_dirrename(sdp, sip, tdp, namep, namlen,
					 tip, &slot);
			}
			macfs_iput(tip);
			break;
		}
	} else {
		/*
		 * The entry does not exist. Check write permission in
		 * directory to see if entry can be created.
		 */
		if (error = macfs_iaccess(tdp, IWRITE)) {
			goto out;
		}
		/*
		 * When renaming an existing entry to a new entry, and 
	 	 * the parent directory is "sticky", then the user must
		 * own the parent directory or the file to be renamed, or
		 * the user must be root.
		 */
		if (op == DE_RENAME && (sdp->o_mode & ISVTX) && u.u_uid != 0 &&
			u.u_uid != sdp->o_uid && u.u_uid != sip->o_uid) {
			error = EPERM;
			goto out;
		}

		if (op == DE_CREATE) {
			/*
			 * make a new inode and directory as required
			 */
			error = macfs_dirmakeinode(tdp, &sip, vap, &slot);
			if (error) {
				goto out;
			}
		}
		error = macfs_diraddentry(tdp, namep, namlen,
					  &slot, sip, sdp, vap);
		if (error) {
			if (op == DE_CREATE) {
				/*
				 * unmake the inode we just made
				 */
				sip->o_nlink = 0;
				sip->i_flag |= ICHG;	/* not imark() */
				macfs_irelease(sip);
				sip = NULL;
			}
		} else if (ipp) {
			MAC_ILOCK(sip);
			*ipp = sip;
		} else if (op == DE_CREATE) {
			macfs_irelease(sip);
		}
	}
/*
	    if(o_mode&IFMT == IFDIR)
		error = mcreatefolder(&imp->mrec,
				      tdp->crec.cdrDirRec.dirDirID,
				      name, &newcr);
	    else
		error = mcreatefile(&imp->mrec,
				      tdp->crec.cdrDirRec.dirDirID,
				      name, &newcr);
*/

out:
	if (slot.bp)
		brelse(slot.bp);
#if	MACH_NBC
	if (sip != NULL && sip->o_nlink == 0)
		inode_uncache(MAC_TOV(sip));
#endif	MACH_NBC
	MAC_IUNLOCK(tdp);
	return (error);
}

/*
 * Delete a directory entry
 * If oip is nonzero the entry is checked to make sure it still reflects oip.
 */
/* BOOGER */
int macfs_dirremove(
	register struct macfs_node *dp,
	char			   *namep,
	struct macfs_node	   *oip,
	int			   rmdir)
{
	register struct macfs_directory_record *ep;
	struct direct *pep;
	struct macfs_node *ip;
	int namlen = strlen(namep);	/* length of name */
	struct slot slot;
	int error = 0;

trace(macfs_debug,("Ddirrem %x %s %x ", dp->o_number, namep, rmdir));
	/*
	 * return error when removing . and ..
	 */
	if (namlen == 1 && namep[0] == '.')
		return (EINVAL);
	if (namlen == 2 && namep[0] == '.' && namep[1] == '.')
		return (ENOTEMPTY);

	ip = NULL;
	slot.bp = NULL;
	MAC_ILOCK(dp);
	/*
	 * Check accessiblity of directory.
	 */
	if ((dp->o_mode&IFMT) != IFDIR) {
		error = ENOTDIR;
		goto out;
	}

	/*
	 * Execute access is required to search the directory.
	 * Access for write is interpreted as allowing
	 * deletion of files in the directory.
	 */
	if (error = macfs_iaccess(dp, IEXEC|IWRITE)) {
		goto out;
	}

	slot.status = FOUND;	/* don't need to look for empty slot */
	ip = (struct macfs_node *)0;
	error = macfs_dircheckforname(dp, namep, namlen, &slot, &ip);
	if (error) {
		goto out;
	}
	if (ip == (struct macfs_node *) 0) {
		error = ENOENT;
		goto out;
	}
	if (oip && oip != ip) {
		error = ENOENT;
		goto out;
	}
	/*
	 * If the parent directory is "sticky", then the user must own
	 * either the parent directory or the file to be removed, or the
	 * user must be root. This implements append-only directories.
	 */
	if ((dp->o_mode & ISVTX) && u.u_uid != 0 &&
	    u.u_uid != dp->o_uid && ip->o_uid != u.u_uid) {
		error = EPERM;
		goto out;
	}

	/* There used to be a check here to make sure you are not removing a
	 * a mounted on dir.  This was no longer correct because iget() does
	 * not cross mount points anymore so the the o_dev fields in the inodes
	 * pointed to by ip and dp will never be different.  There does need
	 * to be a check here though, to eliminate the race between mount and 
	 * rmdir (It can also be a race between mount and unlink, if your 
	 * kernel allows you to unlink a directory.)  
	 */

	if (MAC_TOV(ip)->v_vfsmountedhere != (struct vfs *)0) {
		error = EBUSY;
		goto out;
	}


	/*
	 * If the inode being removed is a directory, we must be
	 * sure it only has entries "." and "..".
	 */
	if (rmdir && (ip->o_mode&IFMT) == IFDIR) {
		if (!macfs_dirempty(ip)) {
			error = ENOTEMPTY;
			goto out;
		}
	}
	/*
	 * Mark entry as deleted.  Should eventually reclaim
	 * room, e.g. if last and only entry in a block. xxx
	 */
	ep = slot.ep;
	ep->name[0] = MAC_NAME_DELETED;
	bwrite(slot.bp);
	slot.bp = NULL;

	/*
	 * Remove the cache'd entry, if any.
	 * Do it after we let the buffer go,
	 * because iupdat might need it.
	 */
	dnlc_remove(MAC_TOV(dp), namep);

	if (rmdir && (ip->o_mode & IFMT) == IFDIR) {
		dnlc_remove(MAC_TOV(ip), ".");
		dnlc_remove(MAC_TOV(ip), "..");
	}
	/*
	 * Remove the inode iff we must.
	 */
	if (oip == (struct macfs_node *)0)
		ip->o_nlink = 0;
out:
	if (slot.bp)
		brelse(slot.bp);
	if (ip) {
		if (ip->o_nlink == 0) {
#if	MACH_NBC
			/*
			 * Remove it from the cache.
			 */
			inode_uncache(MAC_TOV(ip));
#endif	MACH_NBC
			macfs_igone(ip);
		}
		macfs_iput(ip);
	}
#if	MACH_NBC
	if (dp->o_nlink == 0)
		inode_uncache(MAC_TOV(dp));
#endif	MACH_NBC
	MAC_IUNLOCK(dp);
	return (error);

}

/*
 * Check name is ok for a mac file/directory
 */
int macfs_dirbadname(
	register char	*nm,
	register int	l,
	register int	dotrans)
{
	register char c;

trace(macfs_debug,("Ddirbadn %s ", nm));
	if (dotrans & MACFSMNT_MDOT_TR) {
		if ((l > 12) || (l > 8 && nm[8] != '.'))
			return -1;
	} else
		if (l > 11)
			return -1;
	
	/* Like Unix, no nulls or high bits (mumble..) */
	while (l--) {
		c = *nm++;
		if ((c == 0) || (c & 0x80)) {
			return -1;
		}
	}
	return (*nm);
}

/*
 * Perform consistency checks on a directory
 */
int macfs_dirmangled(
	struct macfs_node	*dp,
	struct macfs_directory_record *ep,
	off_t			block_offset)
{
	/* If we had some info available. But. */
	return 0;
}

/*
 * Complain about a bad directory
 */
macfs_dirbad(
	struct macfs_node	*ip,
	char			*how,
	int			offset)
{
	printf("%s: bad dir ino x%x at offset %d: %s\n",
	    ip->i_mnt->im_path, ip->o_number, offset, how);
}

/*
 * Check if the source directory SIP is in the path
 * of the target directory TDP.
 * Target is supplied locked, source is unlocked.
 * The target is always relocked before returning.
 */
int macfs_dircheckpath(
	struct macfs_node       *sip,
	struct macfs_node	*tdp)
{
	struct buf			*bp;
	struct macfs_directory_record	*dirp;
	struct macfs_node		*ip;
#define RENAME_IN_PROGRESS	0x01
#define RENAME_WAITING		0x02
	static int serialize_flag = 0;
	int error = 0;

trace(macfs_debug,("Ddirckp %x %x ", sip->o_number, tdp->o_number));
	/*
	 * If two renames of directories were in progress at once, the partially
	 * completed work of one dircheckpath could be invalidated by the other
	 * rename.  To avoid this, all directory renames in the system are
	 * serialized.
	 */
	while (serialize_flag & RENAME_IN_PROGRESS) {
		serialize_flag |= RENAME_WAITING;
		(void) sleep((caddr_t) &serialize_flag, PINOD);
	}
	serialize_flag = RENAME_IN_PROGRESS;
	ip = tdp;
	if (ip->o_number == sip->o_number) {
		error = EINVAL;
		goto out;
	}
	if (ip->o_number == MAC_ROOTINO)
		goto out;

	bp = 0;
	for (;;) {
		ino_t	dotdot_ino;

		if (((ip->o_mode&IFMT) != IFDIR) ||
		    (ip->o_nlink == 0) ||
		    (ip->i_size < sizeof(struct macfs_directory_record))) {
			macfs_dirbad(ip, "bad size, unlinked or not dir", 0);
			error = ENOTDIR;
			break;
		}
		if (bp) {
			brelse(bp);
		}
		error = macfs_blkatoff(ip, (off_t)0, (char **)&dirp, &bp);
		if (error)
			break;

		/*
		 * .. is guaranteed to be there, second entry.
		 */
		dirp++;
		if (dirp->name[0] != '.' || dirp->name[1] != '.' ||
		    dirp->name[2] != ' ' || dirp->ext[0] != ' ') {
			macfs_dirbad(ip, "mangled .. entry", 0);
			error = ENOTDIR;
			break;
		}

		ino_dtou(dotdot_ino,dirp);

		if (dotdot_ino == MAC_ROOTINO)
			break;

		if (dotdot_ino == sip->o_number) {
			error = EINVAL;
			break;
		}
		if (ip != tdp) {
			macfs_iput(ip);		/* XXX should be zfree ? XXX */
		} else {
			MAC_IUNLOCK(ip);
		}
		/*
		 * o_dev, extent and ih_fs are still valid after macfs_iput
		 * This is a race to get ".." just like dirlook.
		 * But I dont like it.
		 */
		macfs_iget( ip, dotdot_ino, &ip, dirp, sizeof(*dirp));
		if (error)
			break;
	}
	if (bp)
		brelse(bp);
out:
	if (ip) {
		if (ip != tdp) {
			macfs_iput(ip);
			/*
			 * Relock target and make sure it has not gone away
			 * while it was unlocked.
			 */
			MAC_ILOCK(tdp);
			if ((error == 0) && (tdp->o_nlink == 0)) {
				error = ENOENT;
			}
		}
	}
	/*
	 * unserialize
	 */
	{
		register int t = serialize_flag;

		serialize_flag = 0;
		if (t & RENAME_WAITING)
			wakeup((caddr_t) &serialize_flag);
	}

	return (error);
}

/*
 * Check for the existence of a slot to make a directory entry.
 * On successful return *ipp points at the (locked) inode found.
 * The target directory inode (tdp) is supplied locked.
 * This may not be used on "." or "..", but aliases of "." are ok.
 */
int macfs_dircheckforname(
	struct macfs_node	*tdp,	/* inode of directory being checked */
	char			*namep,	/* name we're checking for */
	int			namlen,	/* length of name */
	struct slot		*slotp,	/* slot structure */
	struct macfs_node	**ipp)	/* return inode if we find one */
{
	int dirsize;			/* size of the directory */
	struct buf *bp;			/* pointer to directory block */
	register int entryoffsetinblock;/* offset of ep in bp's buffer */
	register struct macfs_directory_record *ep;	/* directory entry */
	register int offset;		/* offset in the directory */
	struct macfs_node *ip;
	int error, dotrans;

trace(macfs_debug,("Ddirckfn %x %s ", tdp->o_number, namep));
	bp = NULL;
	entryoffsetinblock = 0;
	dotrans = tdp->i_mnt->im_flags & MACFSMNT_TRANS;
	/*
	 * Must search whole directory
	 */
	dirsize = tdp->i_size;
	offset = 0;
	while (offset < dirsize) {
		/*
		 * If offset is on a block boundary,
		 * read the next directory block.
		 * Release previous if it exists.
		 */
		if (macfs_blkoff(tdp->i_mnt, offset) == 0) {
			if (bp != NULL)
				brelse(bp);
			if (error = macfs_blkatoff(tdp, offset,
						 (char **)0, &bp))
				return error;
			entryoffsetinblock = 0;
		}
		ep = (struct macfs_directory_record *)
			(bp->b_un.b_addr + entryoffsetinblock);

		if (macfs_dirmangled(tdp, ep, entryoffsetinblock))
			goto next_one;

		/*
		 * If an appropriate sized slot has not yet been found,
		 * check to see if this one is available.
		 */
		{
			register unsigned int	c = ep->name[0];

			if (c == MAC_NAME_EMPTY || c == MAC_NAME_DELETED) {
			    if (slotp->status != FOUND) {

				slotp->status = FOUND;
				slotp->offset = offset;
			    }
			    goto next_one;
			}
		}
		/*
		 * Check for a name match.
		 */
		if (macfsfncmp(namep, namlen, ep, dotrans)) {
			ino_t	ep_ino;

			tdp->o_diroff = offset;
			ino_dtou(ep_ino,ep);

			if (tdp->o_number == ep_ino) {
				*ipp = tdp;	/* we want ourself, ie "." */
				VN_HOLD(MAC_TOV(tdp));
			} else {
				error = macfs_iget(tdp, ep_ino, &ip, ep, offset);
				if (error) {
					brelse(bp);
					return (error);
				}
				*ipp = ip;
			}
			slotp->status = EXIST;
			slotp->offset = offset;
			slotp->bp = bp;
			slotp->ep = ep;
			return (0);
		}
next_one:
		offset += sizeof *ep;
		entryoffsetinblock += sizeof *ep;
	}
	if (bp) {
		brelse(bp);
	}
	if (slotp->status == NONE) {
		slotp->offset = offset;
	}
	*ipp = (struct macfs_node *)0;
	return (0);
}

/*
 * Rename the entry in the directory TDP so that
 * it points to SIP instead of TIP.
 */
int macfs_dirrename(
	struct macfs_node	*sdp,	/* parent directory of source */
	struct macfs_node	*sip,	/* source inode */
	struct macfs_node	*tdp,	/* parent directory of target */
	char			*namep,	/* entry we are trying to change */
	int			namlen,	/* length of entry string */
	struct macfs_node	*tip,	/* locked target inode */
	struct slot		*slotp)	/* slot for entry */
{
	int error;
	int doingdirectory;
	int parentdifferent;

trace(macfs_debug,("Ddirren %x %s %x ", tdp->o_number, namep, tip->o_number));
	/*
	 * Must have write permission to rewrite target entry.
	 */
	if (error = macfs_iaccess(tdp, IWRITE)) {
		return (error);
	}
	/*
	 * If the parent directory is "sticky", then the user must
	 * own the parent directory or the file to be renamed 
	 * and the target, or the user must be root. 
	 * This implements append-only directories.
	 */
	if ((sdp->o_mode & ISVTX) && u.u_uid != 0 &&
	    u.u_uid != sdp->o_uid && (u.u_uid != sip->o_uid ||
	    u.u_uid != tip->o_uid)) {
		return (EPERM);
	}

	doingdirectory = ((sip->o_mode & IFMT) == IFDIR);
	parentdifferent = (sdp != tdp);
	/*
	 * Check that everything is on the same filesystem.
	 */
	if ((tip->i_vnode.v_vfsp != tdp->i_vnode.v_vfsp) ||
	    (tip->i_vnode.v_vfsp != sip->i_vnode.v_vfsp)) {
		return (EXDEV);		/* XXX archaic */
	}
	/*
	 * Ensure source and target are compatible
	 * (both directories or both not directories).
	 * If target is a directory it must be empty
	 * and have no links to it.
	 */
	if ((tip->o_mode & IFMT) == IFDIR) {
		/*
		 * Target is a dir. Source must be a dir and
		 * target must be empty.
		 */
		if (!doingdirectory) {
			error = ENOTDIR;
			goto bad;
		}
		if (!macfs_dirempty(tip)) {
			error = ENOTEMPTY;
			goto bad;
		}
	} else if (doingdirectory) {
		/*
		 * Source is a dir and target is not.
		 */
		error = ENOTDIR;
		goto bad;
	}
	/*
	 * Rewrite the inode pointer for target name entry
	 * from the target inode (ip) to the source inode (sip).
	 * This prevents the target entry from disappearing
	 * during a crash. Mark the directory inode to reflect the changes.
	 */
	dnlc_remove(MAC_TOV(tdp), namep);
	sip->macfs_diroff = slotp->offset;
	sip->macfs_dirextent = tdp->extent;
	(void) macfs_itdiren(sip,slotp->ep, sip->i_mnt->im_flags & MACFSMNT_MUSER);
	dnlc_enter(MAC_TOV(tdp), namep, MAC_TOV(sip), NOCRED);
	bwrite(slotp->bp);
	slotp->bp = NULL;

	/*
	 * The target inode is gone.
	 * Fix the ".." entry in sip to point to dp.
	 * This is done after the new entry is on the disk.
	 */
	tip->o_nlink--;
	if (doingdirectory) {
		/*
		 * Decrement target link count once more if it was a directory.
		 */
		if (--tip->o_nlink != 0) {
			panic("direnter: target directory link count");
		}
		/*
		 * Renaming a directory with the parent different requires
		 * ".." to be rewritten. The window is still there for ".."
		 * to be inconsistent, but this is unavoidable.
		 */
		if (parentdifferent) {
			error = macfs_dirfixdotdot(sip, sdp, tdp);
			if (error) {
				goto bad;
			}
		}
	}
bad:
	if (tip->o_nlink != 0) printf("dirren link %d\n", tip->o_nlink);
#if	MACH_NBC
#else	MACH_NBC
	if ((tip->i_flag & ITEXT) == ITEXT)
#endif	MACH_NBC
		inode_uncache(MAC_TOV(tip));
	macfs_igone(tip);
	return (error);
}

/*
 * Enter the file sip in the directory tdp with name namep.
 */
int macfs_diraddentry(
	struct macfs_node	*tdp,
	char			*namep,
	int			namlen,
	struct slot		*slotp,
	struct macfs_node	*sip,
	struct macfs_node	*sdp,
	struct vattr		*vap)
{
	int error;
	register struct macfs_directory_record	*ep;

trace(macfs_debug,("Ddiradd %x %s %x ", tdp->o_number, namep, sip->o_number));
	/*
	 * Prepare a new entry. Grabs a buffer in slotp->bp.
	 */
	error = macfs_dirprepareentry(tdp, slotp);
	if (error) {
		return (error);
	}
	/*
	 * Check inode to be linked to see if it is in the
	 * same filesystem.
	 */
	if (tdp->i_vnode.v_vfsp != sip->i_vnode.v_vfsp) {
		error = EXDEV;
		goto bad;
	}
	if ((sip->o_mode & IFMT) == IFDIR) {
		error = macfs_dirfixdotdot(sip, sdp, tdp);
		if (error) {
			goto bad;
		}
	}
	/*
	 * Fill in entry data
	 */
	ep = slotp->ep;
	macfsfntomac(ep, namep, namlen, 
		     (tdp->i_mnt->im_flags & MACFSMNT_TRANS));
	macfs_macdate(sip->o_ctime,ep->time,ep->date);
	ep->attr = sip->attr;
	macfs_itdiren(sip,ep,tdp->i_mnt->im_flags & MACFSMNT_MUSER);
	
	dnlc_enter(ITOV(tdp), namep, ITOV(sip), NOCRED);
	/*
	 * Write out the directory entry.
	 */
	bwrite(slotp->bp);
	slotp->bp = NULL;
	/*
	 * Mark the directory inode to reflect the changes.
	 */
	tdp->o_diroff = 0;
	return (0);

bad:
	/*
	 * Clear out entry prepared by dirprepareent.
	 */
	slotp->ep->name[0] = MAC_NAME_DELETED;
	bwrite(slotp->bp);
	slotp->bp = NULL;
	return (error);
}

/*
 * Is directory IP empty ?
 */
int macfs_dirempty(
	struct macfs_node	*ip)
{
	struct direct dirent;
	int macfs_offset;
	int entryoffsetinblock;
	int error = 0, dotrans;
	int endsearch;
	struct macfs_directory_record *ep;
	struct macfs_mount *imp;
	struct buf *bp = NULL;

trace(macfs_debug,("Ddirempty %x ", ip->o_number));

	if (ip->o_number == MAC_ROOTINO)
		return 0;	/* sanity */

	imp = ip->i_mnt;
	dotrans = imp->im_flags & MACFSMNT_TRANS;
	macfs_offset = entryoffsetinblock = 0;
	endsearch = ip->i_size;

	while (macfs_offset < endsearch) {
		/*
		 * If offset is on a block boundary,
		 * read the next directory block.
		 * Release previous if it exists.
		 */

		if (macfs_blkoff(imp, macfs_offset) == 0) {
			if (bp != NULL)
				brelse(bp);
			if (error = macfs_blkatoff(ip, macfs_offset,
						 (char **)0, &bp))
				return (0);
			entryoffsetinblock = 0;
		}
		/*
		 * Get pointer to next entry.
		 */

		ep = (struct macfs_directory_record *)
			(bp->b_un.b_addr + entryoffsetinblock);

		/*
		 * Special cases
		 */
		if (ep->name[0] == MAC_NAME_EMPTY)
			goto next_one;

		if (ep->name[0] == MAC_NAME_DELETED)
			goto next_one;
		
		/*
		 * Translate the name
		 */
		if (macfsfntrans(ep, dirent.d_name, &dirent.d_namlen, dotrans))
			goto out; /* ?? */

		/*
		 * "empty" means only . and ..
		 */
		if (dirent.d_namlen > 2)
			goto out;
		if (dirent.d_name[0] != '.')
			goto out;
		if ( ! (dirent.d_namlen == 1 || dirent.d_name[1] == '.') )
			goto out;
next_one:
		macfs_offset += sizeof(struct macfs_directory_record);
		entryoffsetinblock += sizeof(struct macfs_directory_record);
	}
	error = 1;
out:
	if (bp)
		brelse (bp);
	return error;
}

/*
 * Make an empty directory in the inode ip.
 */
struct macfs_directory_record macfs_mastertemplate[2] =
{ { ".       ", "   ", MAC_ATTR_DIR, },
  { "..      ", "   ", MAC_ATTR_DIR, } };

macfs_dirmakedirect(
	register struct macfs_node *ip,		/* new directory */
	register struct macfs_node *dp)		/* parent directory */
{
	register struct macfs_directory_record *dirp;
	struct buf *bp;
	struct macfs_mount	*imp;
	daddr_t bn;
	int error, size;

trace(macfs_debug,("\nmacfs_dirmakedirect %x %x ", dp->o_number, ip->o_number));
	/*
	 * Allocate space for the directory we're creating.
	 */
	imp = ip->i_mnt;
	size = macfs_blksize(imp,ip,0);
	error = macfs_iextend(ip, size, 0);
	if (error)
		return error;
	imark(ip, IUPD|ICHG);

	/*
	 * Get and clear the directory's first block
	 */
	bn = macfs_bmap(ip, (daddr_t)0);
	bp = getblk(ip->o_devvp, bn, size);
	clrbuf(bp);
	dirp = (struct macfs_directory_record *)bp->b_un.b_addr;

	/*
	 * Now initialize the directory we're creating with the
	 * "." and ".." entries.  Then write it out to disk.
	 */
	dirp[0] = macfs_mastertemplate[0];
	ino_utod(ip->o_number,&dirp[0]);
	dirp[1] = macfs_mastertemplate[1];
	ino_utod(dp->o_number,&dirp[1]);
	bwrite(bp);
	return (0);
}

int macfs_fiddebug = 0;

int macfs_fid(
	struct vnode *vp,
	struct fid **fidpp)
{
	register struct ifid *ifid;
	struct macfs_node	*ip = VTO_MAC(vp);

trace(macfs_fiddebug,("macfs_fid %x (%x) ", ip, ip->macfs_dirextent));
	ZALLOC(macfs_fid_zone, ifid, struct ifid *);
	bzero((caddr_t)ifid, sizeof(struct ifid));
	ifid->ifid_len = sizeof(struct ifid) - (sizeof(struct fid) - MAXFIDSZ);
	ifid->ifid_ino = ip->o_number;
	ifid->ifid_dirext = ip->macfs_dirextent;
	ifid->ifid_off = ip->macfs_diroff;

	*fidpp = (struct fid *)ifid;
	return (0);
}

int macfs_freefid(
	struct vnode *vp,
	struct fid *fidp)
{
trace(macfs_fiddebug,("macfs_freefid %x\n", vp));
	ZFREE(macfs_fid_zone, fidp);
	return (0);
}

/*
 * File handle to vnode
 *
 * Have to be really careful about stale file handles:
 * - check that the inode number is in range
 * - call iget() to get the locked inode
 * - check for an unallocated inode (i_mode == 0)
 * - check that the generation number matches
 */
int macfs_fhtovp(
	register struct vfs *vfsp,
	struct vnode	**vpp,
	struct fid	*fhp)
{
	struct macfs_node tvp, *nip;
	struct macfs_directory_record *dirp;
	int error, off, bsiz;
	daddr_t block;
	register struct ifid *ifhp;
	struct macfs_mount	*macfsmp;
	struct buf	*bp;

	ifhp = (struct ifid *)fhp;
	macfsmp = VFSTOMACFS (vfsp);

trace(macfs_fiddebug,("macfs_fhtovp %x %x %x %x\n", 
		ifhp->ifid_len, ifhp->ifid_ino,
		ifhp->ifid_dirext, ifhp->ifid_off));

	*vpp = NULL;
	if (ifhp->ifid_ino == MAC_ROOTINO)
		return macfs_root(vfsp, vpp);


	off = ifhp->ifid_off;
	if (ifhp->ifid_dirext >= macfsmp->im_size)
		return (EINVAL);

	bsiz = macfsmp->im_bsize;
	if ((off & (bsiz-1)) + sizeof (struct macfs_directory_record) >= bsiz)
		return (EINVAL);

	/*
	 * Get to the directory record for the inode
	 */
	tvp.i_vnode.ih_fs = vfsp;
	tvp.o_number = ifhp->ifid_ino; /* not really, but != rootino */
	tvp.i_mnt = macfsmp;
	tvp.last_lbn = -1;
	tvp.o_dev = macfsmp->im_dev;
	tvp.extent = ifhp->ifid_dirext;
	tvp.macfs_diroff = off;

	bp = bread (	macfsmp->im_devvp,
			macfs_bmap(&tvp, macfs_lblkno(macfsmp,off)),
			bsiz);

	dirp = (struct macfs_directory_record *)(bp->b_un.b_addr + (off & (bsiz-1)));

	error = macfs_iget(&tvp, ifhp->ifid_ino, &nip, dirp, off);
	if (error == 0) {
		MAC_IUNLOCK(nip);
		*vpp = MAC_TOV(nip);
	}
out:
	brelse (bp);
	return (error);
}

/* BOOGER */
int macfs_itdiren(
	struct macfs_node	*ip,
	struct macfs_directory_record *dirp,
	int			extensions)
{
	register int tmp, upd = 0;;

	tmp = macfsnum_32(dirp->size);
	if (ip->attr & MAC_ATTR_DIR) {
		/* sanity for MAC */
		if (tmp != 0) {
			mkmacfsnum_32(dirp->size,0);
			upd++;
		}
	} else {
		if (tmp != ip->i_size) {
			upd++;
			tmp = ip->i_size;
			mkmacfsnum_32(dirp->size,tmp);
		}
	}

	ino_dtou(tmp,dirp);
	if (tmp != ip->o_number) {
		/* should only happen on file creation */
		upd++;
		ino_utod(ip->o_number,dirp);
	}

	if (!extensions)
		return upd;

	tmp = macfsnum_16(dirp->res_ext.mnt_ext.mode);
	if (tmp != ip->o_mode) {
		upd++;
		tmp = ip->o_mode;
		mkmacfsnum_16(dirp->res_ext.mnt_ext.mode, tmp);
	}

	tmp = macfsnum_16(dirp->res_ext.mnt_ext.gid);
	if (tmp != ip->o_gid) {
		upd++;
		tmp = ip->o_gid;
		mkmacfsnum_16(dirp->res_ext.mnt_ext.gid, tmp);
	}

	tmp = macfsnum_16(dirp->res_ext.mnt_ext.uid);
	if (tmp != ip->o_uid) {
		upd++;
		tmp = ip->o_uid;
		mkmacfsnum_16(dirp->res_ext.mnt_ext.uid, tmp);
	}

	/*
	 * For now, put the device numbers in the extended name area
	 */
	if (ISVDEV(ip->o_mode & VFMT)) {
		tmp = macfsnum_32(dirp->res_ext.mnt_ext.name2);
		if (tmp != ip->i_vnode.v_rdev) {
			upd++;
			tmp = ip->i_vnode.v_rdev;
			mkmacfsnum_32(dirp->res_ext.mnt_ext.name2, tmp);
		}
	}
	return upd;	
}

/* BOOGER */
int macfs_itrunc(struct macfs_node *ip, int size)
{
mgrowfile(mr, crec, newnumblock) returns

MGROWDOESNTTRUNC
MGROWGREWTOWRONGSIZE
MGROWNOTENOUGHSPACE

    if(mtruncatefile(ip->mr, &ip->crec, newsize))
	return(error);
}
