/*
The vfs structure:

struct vfs {
	struct vfs	*vfs_next;		// next vfs in vfs list
	struct vfsops	*vfs_op;		// operations on vfs
	struct vnode	*vfs_vnodecovered;	// vnode we mounted on
	int		vfs_flag;		// flags
	int		vfs_bsize;		// native block size
	fsid_t		vfs_fsid;		// file system id
	u_short		vfs_exroot;		// exported fs uid 0 mapping
	short		vfs_exflags;		// exported fs flags
	caddr_t		vfs_data;		// private data
};

The filesystem operations are:

struct vfsops {
	int	(*vfs_mount)();		// mount file system(vfs*, path, data)
	int	(*vfs_unmount)();	// unmount file system(vfs*)
	int	(*vfs_root)();		// get root vnode(vfs*, vnode**)
	int	(*vfs_statfs)();	// get fs statistics(vfs*, statfs*)
	int	(*vfs_sync)();		// flush fs buffers(vfs*)
	int	(*vfs_vget)();		// vnode from fid(vfs*, vnode**, fid*)
};

The vnode operations are

struct vnodeops {
	int	(*vn_open)();       // vnode**, flag, cred
	int	(*vn_close)();      // vnode*, flag, cred
	int	(*vn_rdwr)();       // vnode*, uio*, rw, ioflag, cred
	int	(*vn_ioctl)();      // vnode*, com, data, flag
	int	(*vn_select)();     // vnode*, which, cred UNSUPPORTED
	int	(*vn_getattr)();    // vnode*, vattr*, cred
	int	(*vn_setattr)();    // vnode*, vattr*, cred
	int	(*vn_access)();     // vnode*, mode, cred (access check)
0	int	(*vn_lookup)();     // vnode*, name, vnode**, cred
	int	(*vn_create)();     // vnode*, name, vattr, excl,
                                       // mode, vnode**, cred
	int	(*vn_remove)();     // vnode*, name, cred
	int	(*vn_link)();       // vnode*, vnode*, name, cred
	int	(*vn_rename)();     // vnode*, name, vnode*, name, cred
	int	(*vn_mkdir)();      // vnode*, name, vattr, vnode**, cred
	int	(*vn_rmdir)();      // vnode*, name, cred
	int	(*vn_readdir)();    // vnode*, uio, cred
	int	(*vn_symlink)();    // vnode*, name, vattr, name, cred
	int	(*vn_readlink)();   // vnode*, uio*, cred
	int	(*vn_fsync)();      // vnode*, cred
	int	(*vn_inactive)();   // vnode*, cred
	int	(*vn_bmap)();       // vnode*, FsBlock#, vnode**, DevBlock#
	int	(*vn_strategy)();   // Ignored
	int	(*vn_bread)();      // vnode*, LogBlock, buf**
	int	(*vn_brelse)();     // vnode*, buf*
	int	(*vn_lockctl)();    // vnode*, flock*, cmd, cred
	int	(*vn_fid)();        // vnode*, fid**
	int	(*vn_page_read)();  // vnode*, buf, size, vm_offset_t, cred
	int	(*vn_page_write)(); // vnode*, buf, size, vm_offset_t,
                                       // cred, initializing
	int	(*vn_read1dir)();   // same as readdir
	int	(*vn_freefid)();    // vnode*, fid*
};
*/